/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "SpringDashPot.h"
using namespace std;

#include <vector>
#include <math.h>

#include <QUB_Heap.h>
#include "qubopt.h"
#include "IdlCond.h"
#include "LoadDataCB_Pager.h"
#include "qub_reportstream.h"


using namespace std;

#define IDLCOND_PAGESIZE (2 << 12)

QUB_Tree IdlCond1DataSet( QUB_Tree dataSet, QUB_Tree config, QTR_NodeFptr fillDataCB, 
						  ostream& report, QTR_NodeFptr pctCB );
/*
Properties
{
	Known Amps = 0.0 5.0
	AddDeltaAmps = 1
	Delta = 0.25
	MinDur_MS = 5.0
	Latency_MS = 0.0
	SpringDashPot = 0
	SDP Damping = 0.6
	SDP Freq = 1100.0
	Tension = 0
	OnThreshold = 0.5
	RestingTension = 0.0
	CollapsedTension = 0.0
	TensionPressure = 1.0
	TauRising = 1.0
	TauCollapsing = 0.01
	TauRecovering = 0.1
}

DataSet
{
	sampling=0.1
	Scaling=32.1
	ActiveChannel=1
	Channels
	{
		Scaling=0.1
	}
	Segment=0 4999
	{
		start=0#ms

		# creating:
		amp = 0.0 1.0
		sd = 0.1 0.1
		DwellCount = 3
		(Classes	Durations	Firsts	Lasts
		 0			3.2			0		31
		 1			.1			32		32
		 0			.9			33		41)
	}
}
*/

int sample_const(double a, int n, double *data, int ndata, double *xdata, int nxdata)
{
	int overflow = max(0, n-ndata);
	int overbite = min(nxdata, overflow);
	for (int i=0; i<n-overflow; ++i)
		data[i] = a;
	if ( xdata )
		for (int i=0; i<overbite; ++i)
			xdata[i] = a;
	return overflow;
}

int sample_exp(double a, double b, double tau, int n, double sampling, double *data, int ndata, double *xdata, int nxdata)
{
	int overflow = max(0, n-ndata);
	int overbite = min(nxdata, overflow);
	// a + (b-a)*(1 - exp(-(x-x0) / Tau))
	//   -x / Tau; -i*sampling / Tau; i*(-sampling/Tau)
	// a + (b-a)*(1 - exp(i*(-sampling/tau))
	double bma = b-a;
	double msot = -sampling / tau;
	for (int i=0; i<n-overflow; ++i)
		data[i] = a + bma*(1.0 - exp(i*msot));
	if ( xdata )
		for (int i=0; i<overbite; ++i)
			xdata[i] = a + bma*(1.0 - exp((i+n-overbite)*msot));
	return overflow;
}

// to use, define bool flip = onThreshold < 0
#define ABOVE_THRESH(amp)  ((flip) ? ((amp) <= onThreshold) : ((amp) >= onThreshold))

class TensionPulses_Pager : public QUB_Pager
{
public:
	QUB_Tree idlSeg;
	int idl_ix;
	double *xdata;
	double onThreshold, resting, collapsed, rising, collapsing, recovering, ToP;
	double sampling;
	int overflow;

	TensionPulses_Pager(QUB_Tree segNode, QUB_Tree config, QTR_NodeFptr fillDataCB, 
						  ostream& report, QTR_NodeFptr pctCB)
	: QUB_Pager(((int *) segNode.data())[1] - ((int *) segNode.data())[0] + 1,
				IDLCOND_PAGESIZE, - IDLCOND_PAGESIZE, NULL),
	  onThreshold( config["OnThreshold"].dataAsDouble(0.5) ),
	  resting( config["RestingTension"].dataAsDouble(0.0) ),
	  collapsed( config["CollapsedTension"].dataAsDouble(0.0) ),
	  rising( config["TauRising"].dataAsDouble(1.0) ),
	  collapsing( config["TauCollapsing"].dataAsDouble(0.01) ),
	  recovering( config["TauRecovering"].dataAsDouble(0.1) ),
	  ToP( config["TensionPressure"].dataAsDouble(1.0) )
	{
		// set up a one-segment dataset
		QUB_Tree ds = QUB_Tree::Create("DataSet");
		QUB_Tree seg = ds["Segment"];
		ds.appendClone( segNode.parent()["FileName"] );
		ds.appendClone( segNode.parent()["ADChannelCount"] );
		ds.appendClone( segNode.parent()["ActiveChannel"] );
		ds.appendClone( segNode.parent()["sampling"] );
		sampling = segNode.parent()["sampling"].dataAsDouble(1.0);

		QUB_Tree procData = * ( segNode.parent().find("ProcessData") );
		if ( ! procData.isNull() )
			ds.appendClone( procData );

		seg.setNumData( QTR_TYPE_INT, 2, 1 );
		memcpy(seg.data(), segNode.data(), 2*sizeof(int));
		if ( dataCount < pageSize )
			pageSize = dataCount;
		pageStart = - pageSize;

		QUB_TreeMonoIter ix = segNode.find("Index");
		if ( ! (*ix).isNull() )
			seg.appendClone( (*ix) );

		// run non-tension processing -- same config tree to share stop flag
		config["Tension"].setData(QTR_TYPE_INT, 0);
		config["MinDur_MS"].dataAs(0, 0, (double)0.0) *= 1e-1;
		QUB_Tree result = IdlCond1DataSet(ds, config, fillDataCB, report, pctCB);
		config["Tension"].setData(QTR_TYPE_INT, 1);
		config["MinDur_MS"].dataAs(0, 0, (double)0.0) *= 1e1;
		idlSeg = result["Segment"];

		applyThreshold();
		idl_ix = 0;

		// output arrays
		data = new double[IDLCOND_PAGESIZE];
		xdata = new double[IDLCOND_PAGESIZE];
		overflow = -1;
	}

	virtual ~TensionPulses_Pager()
	{
		delete [] data;
		delete [] xdata;
	}

	void applyThreshold()
	{
		if ( ! idlSeg["DwellCount"].data() )
			return;
		int& nevent = * (int *) idlSeg["DwellCount"].data();
		double *amp = (double *) idlSeg["amp"].data();
		int *idwt = (int *) idlSeg["Classes"].data();
		int *sdwt = (int *) idlSeg["Firsts"].data();
		int *edwt = (int *) idlSeg["Lasts"].data();
		if ( ! (nevent && amp && idwt && sdwt && edwt) )
			return;

		bool flip = onThreshold < 0;
		bool a_on = ABOVE_THRESH(amp[idwt[0]]);
		int a = 0; int b; // left and right reading pointers
		int j = 0; // next output event
		while ( a < nevent ) {
			for ( b = a+1; b < nevent; ++b ) {
				if ( a_on != ABOVE_THRESH(amp[idwt[b]]) ) {
					a_on = ! a_on;
					break;
				}
			}
			// output one event for a..b-1
			// with mode amplitude
			vector<int> counts(idlSeg["amp"].dataCount(), 0);
			for ( int i=a; i<b; ++i )
				++counts[idwt[i]];
			int mode = 0;
			int modeCount = 0;
			for ( int i=0; i<counts.size(); ++i ) {
				if ( counts[i] > modeCount ) {
					modeCount = counts[i];
					mode = i;
				}
			}
			// just to point out, we are not updating idlSeg["Durations"]
			idwt[j] = mode;
			sdwt[j] = sdwt[a];
			edwt[j] = edwt[b-1];
			++j;

			a = b;
		}
		nevent = j;
	}

	void nextPage()
	{
		pageStart += pageSize;

		if ( ! idlSeg["DwellCount"].data() )
			return;
		int& nevent = * (int *) idlSeg["DwellCount"].data();
		double *amp = (double *) idlSeg["amp"].data();
		int *idwt = (int *) idlSeg["Classes"].data();
		int *sdwt = (int *) idlSeg["Firsts"].data();
		int *edwt = (int *) idlSeg["Lasts"].data();
		if ( ! (nevent && amp && idwt && sdwt && edwt) )
			return;

		int segFirst = sdwt[0];
		int segLast = edwt[nevent-1];

		swap();
		int pos;
		bool flip = onThreshold < 0;
		bool on = ABOVE_THRESH(amp[idwt[0]]);
		double tension = on ? (amp[idwt[0]] * ToP) : resting;
		int len = edwt[idl_ix]-(segFirst+pageStart)+1;

		if ( overflow >= pageSize ) {
			overflow = -6; // at most one page ahead has been written; re-consider event next page
			return;
		} else if ( overflow == -6 ) {
			pos = 0;
		} else if ( overflow >= 0 ) {
			++idl_ix;
			on = ! on;
			pos = overflow;
		} else {
			overflow = sample_const(tension, len, data, pageSize, xdata, pageSize);
			pos = len;
			if ( ! overflow ) {
				++idl_ix;
				on = ! on;
			}
		}
		while ( (idl_ix < nevent) && (pos < pageSize) ) {
			len = edwt[idl_ix]-(segFirst+pageStart+pos)+1;
			if ( on ) {
				tension = amp[idwt[idl_ix]] * ToP;
				overflow = sample_exp(resting, tension, rising, len, sampling, data+pos, pageSize-pos, xdata, pageSize);
				pos += len;
			} else {
				int lenCollapse = min(len/2, 3*collapsing / sampling);
				overflow = sample_exp(tension, collapsed, collapsing, lenCollapse, sampling, data+pos, pageSize-pos, xdata, pageSize);
				pos += lenCollapse;
				if ( ! overflow ) {
					overflow = sample_exp(collapsed, resting, recovering, len-lenCollapse, sampling, data+pos, pageSize-pos, xdata, pageSize);
				} else {
					sample_exp(collapsed, resting, recovering, len-lenCollapse, sampling, xdata, pageSize-overflow, NULL, 0);
					overflow = min(pageSize, overflow + (len-lenCollapse));
				}
				pos += (len - lenCollapse);
			}
			if ( pos < pageSize ) {
				++idl_ix;
				on = ! on;
			}
		}
		for (int i=pos; i<pageSize; ++i)
			data[i] = 0.0;
	}

	void swap()
	{
		double *tmp = data;
		data = xdata;
		xdata = tmp;
	}
};


// find nearest amp in list of known amps and return whether it's within delta
int FindKnownAmp(vector<double>& known, double amp, double delta, double& ampOut)
{
	assert(known.size() > 0);
	double minDist = abs(amp - known[0]);
	int iMin = 0;
	for ( int i=known.size()-1; i>0; --i ) {
		double dist = abs(amp - known[i]);
		if ( dist < minDist ) {
			minDist = dist;
			iMin = i;
		}
	}
	ampOut = known[iMin];
	return (minDist < delta);
}

class ConsolidInterval
{
public:
	double interval;
	double from, to;

	ConsolidInterval(double i_from, double i_to)
		: interval(i_to - i_from), from(i_from), to(i_to)
	{}
	bool operator<(const ConsolidInterval& other)
	{
		return (interval > other.interval); // to make the "biggest" in the heap be the smallest interval
	}
};

class Consolidator
{
public:
	vector<int>& clss;
	vector<int>& counts;
	map<double, int>& clsOfAmp;
	vector<double>& ampOfCls;
	vector<double>& known;
	int& StopFlag;
	double delta;
	typedef pair<int, vector<int> > SortAmpRec;
	typedef map<double, SortAmpRec> SortAmps;
	SortAmps sortAmps; // map from amp value to (point count, class indices) it represents
	SortAmps knownAmps;
	typedef QUB_Heap<ConsolidInterval> SortInter;
	SortInter sortInter; // heap of intervals between amps, smallest on top

	Consolidator(vector<int>& i_clss, vector<int>& i_counts, map<double, int>& i_clsOfAmp, vector<double>& i_ampOfCls, vector<double>& i_known, int& i_StopFlag)
		: clss(i_clss), counts(i_counts), clsOfAmp(i_clsOfAmp), ampOfCls(i_ampOfCls), known(i_known), StopFlag(i_StopFlag)
	{
		vector<int> countInCls(ampOfCls.size(), 0);
		for ( int i=clss.size()-1; i>=0; --i )
			countInCls[clss[i]] += counts[i];
		for (int i=0; i<(int)ampOfCls.size(); ++i ) {
			if ( find(known.begin(), known.end(), ampOfCls[i]) != known.end() ) {
				SortAmpRec& rec = knownAmps[ampOfCls[i]];
				rec.first = countInCls[i];
				rec.second.push_back(i);
			}
			else {
				SortAmpRec& rec = sortAmps[ampOfCls[i]];
				rec.first = countInCls[i];
				rec.second.push_back(i);
			}
		}
		SortAmps::iterator ai_last = sortAmps.begin();
		SortAmps::iterator ai = ai_last;
		for ( ++ai; ai != sortAmps.end(); ai_last = ai, ++ai ) {
			sortInter.push(ConsolidInterval(ai_last->first, ai->first));
		}
	}

	// this will leave you with a smaller number of classes, in order, separated by at least delta,
	// and it will leave some gaps (clss -1, count 0) in your event record as it joins neighboring events.
	void consolidate(double delta)
	{
		this->delta = delta;
		consolidateAmps();
		renumberClasses();
		consolidateEvents();
	}

	void consolidateAmps() {
		if (sortInter.size() == 0)
			return;
		while (sortInter.size() > 1 && sortInter.first().interval < delta) {
			if ( StopFlag )
				break;

			// we will replace the endpoints (classes) of this too-small interval with one averaged class

			// remove endpoints
			double from = sortInter.first().from;
			double to = sortInter.first().to;
			SortAmpRec fromRec = sortAmps[from];
			SortAmpRec toRec = sortAmps[to];
			sortAmps.erase(from);
			sortAmps.erase(to);

			// add endpoint from weighted average of old endpoints, sum point count, union class indices
			double newAmp = (from * fromRec.first + to * toRec.first) / (fromRec.first + toRec.first);
			SortAmpRec& newAmpRec = sortAmps[newAmp];
			newAmpRec.first = fromRec.first + toRec.first;
			for ( vector<int>::iterator ii = fromRec.second.begin(); ii != fromRec.second.end(); ++ii )
				newAmpRec.second.push_back(*ii);
			for ( vector<int>::iterator ii = toRec.second.begin(); ii != toRec.second.end(); ++ii )
				newAmpRec.second.push_back(*ii);

			// remove interval and neighbors
			// aka remove intervals incident on from and/or to
			// or: pop, remove the interval from "to" and the one to "from"
			sortInter.pop();
			for ( int i=0; i<2; ++i ) { // get one neighboring interval then the other
				for ( SortInter::iterator ii = sortInter.begin(); ii != sortInter.end(); ++ii ) {
					if ( (ii->from == to) || (ii->to == from) ) {
						sortInter.erase(ii);
						break;
					}
				}
			}

			// add intervals incident on new endpoint
			SortAmps::iterator otherAmp = sortAmps.find(newAmp);
			if ( otherAmp != sortAmps.begin() ) {
				--otherAmp;
				sortInter.push(ConsolidInterval(otherAmp->first, newAmp));
			}
			otherAmp = sortAmps.find(newAmp);
			++otherAmp;
			if ( otherAmp != sortAmps.end() ) {
				sortInter.push(ConsolidInterval(newAmp, otherAmp->first));
			}
		}
	}

	void renumberClasses()
	{
		clsOfAmp.clear();
		ampOfCls.clear();
		map<int, int> oldToNewCls;
		int i = 0;
		for ( SortAmps::iterator ai = knownAmps.begin(); ai != knownAmps.end(); ++i, ++ai ) {
			clsOfAmp[ai->first] = i;
			ampOfCls.push_back(ai->first);
			oldToNewCls[ai->second.second[0]] = i;
		}
		for ( SortAmps::iterator ai = sortAmps.begin(); ai != sortAmps.end(); ++i, ++ai ) {
			clsOfAmp[ai->first] = i;
			ampOfCls.push_back(ai->first);
			for ( vector<int>::iterator oci = ai->second.second.begin(); oci != ai->second.second.end(); ++oci )
				oldToNewCls[*oci] = i;
		}

		for ( i=clss.size()-1; i>=0; --i )
			clss[i] = oldToNewCls[clss[i]];
	}

	void consolidateEvents()
	{
		int last = 0;
		for ( int i=1; i<(int)clss.size(); ++i ) {
			if ( clss[i] == clss[last] ) {
				clss[i] = -1;
				counts[last] += counts[i];
				counts[i] = 0;
			}
			else {
				last = i;
			}
		}
	}
};

// build event list by repeated add()
// if you add multiple events in a row with same amp, it joins them
// maxCount helps you break along segment bounds
// ampOfCls is DataSet:Segment:amp

class EventMaker
{
public:
	vector<int> clss;
	vector<int> counts;
	map<double, int> clsOfAmp;
	vector<double> ampOfCls;
	int at;

	EventMaker()
		: at(0)
	{}

	// stage one: add all your events
	void add(double amp, int count)
	{
		int cls = findOrAddCls(amp);
		int ilast = clss.size() - 1;
		if ( ilast >= 0 && cls == clss[ilast] )
			counts[ilast] += count;
		else {
			counts.push_back(count);
			clss.push_back(cls);
		}
	}
	
	// stage two: consolidate
	void consolidate(double delta, vector<double>& known, int& StopFlag)
	{
		Consolidator cons(clss, counts, clsOfAmp, ampOfCls, known, StopFlag);
		cons.consolidate(delta);
	}

	// stage three: pull out finished events
	int nextEvent(int& cls, int& count, int maxCount) // returns count
	{
		while ( at < (int) clss.size() && 0 == counts[at] )
			++at;
		if ( at == clss.size() )
			return (count = 0);

		cls = clss[at];
		count = counts[at];
		if ( count > maxCount ) {
			counts[at] = count - maxCount;
			count = maxCount;
		}
		else {
			++at;
		}
		return count;
	}

	int findOrAddCls(double amp)
	{
		map<double, int>::iterator ii = clsOfAmp.find(amp);
		if ( ii != clsOfAmp.end() )
			return ii->second;

		// (else):
		int cls = ampOfCls.size();
		clsOfAmp[amp] = cls;
		ampOfCls.push_back( amp );
		return cls;
	}
};

// ---------------------
QUB_Tree CreateIdealSegment(QUB_Tree segIn, vector<int>& classes, vector<int>& counts, vector<double>& amps, double sampling/*ms*/)
{
	QUB_Tree segOut = segIn.clone(false);
	segOut.appendClone(segIn["start"]);
	int at = segOut.dataAs(0, 0, 0);

	int nclass = amps.size();
	segOut["amp"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, &(amps[0]));
	QUB_Tree sdNode = segOut["sd"];
	sdNode.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void*)0);
	double *sd = (double *) sdNode.data();
	for ( int i=0; i<nclass; ++i )
		sd[i] = 0.1;

	int ndwt = counts.size();
	segOut["DwellCount"].setData(QTR_TYPE_INT, ndwt);
	QUB_Tree idwtNode = segOut["Classes"];
	idwtNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
	int *idwt = (int *) idwtNode.data();
	QUB_Tree tdwtNode = segOut["Durations"];
	tdwtNode.setNumData(QTR_TYPE_FLOAT, ndwt, 1, (void*)0);
	float *tdwt = (float *) tdwtNode.data();
	QUB_Tree sdwtNode = segOut["Firsts"];
	sdwtNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
	int *sdwt = (int *) sdwtNode.data();
	QUB_Tree edwtNode = segOut["Lasts"];
	edwtNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
	int *edwt = (int *) edwtNode.data();
	
	for ( int i=0; i<ndwt; ++i ) {
		idwt[i] = classes[i];
		tdwt[i] = (float) (sampling * counts[i]);
		sdwt[i] = at;
		at += counts[i];
		edwt[i] = at - 1;
	}
	return segOut;
}

// -------------------------------

QUB_Tree IdlCond1DataSet( QUB_Tree dataSet, QUB_Tree config, QTR_NodeFptr fillDataCB, 
						  ostream& report, QTR_NodeFptr pctCB )
{
	int& StopFlag = ((int*) config["StopFlag"].data())[0];

	QUB_Tree pctNode = QUB_Tree::Create("");
	pctNode.setData( QTR_TYPE_INT, 0 );
	int &pct = * (int *) pctNode.data();
	
	double sampling = dataSet["sampling"].dataAsDouble(1.0);

	vector<double> known;
	QUB_Tree knownNode = config["Known Amps"];
	double *knownData = (double *) knownNode.data();
	for ( int i=0; i<(int)knownNode.dataRows(); ++i )
		known.push_back(knownData[i]);

	bool addDeltas = (0 != config["AddDeltas"].dataAsInt(0));
	double delta = config["Delta"].dataAsDouble(1.0);

	double latency_exact = config["Latency_MS"].dataAsDouble(0.0) / sampling;
	int latency = int(latency_exact);
	if ( latency_exact - latency >= 0.5 )
		++latency;

	QUB_Pager *pager = NULL;
	EventMaker eventMaker;
	double amp;
	int count;

	int nseg = 0;
	for (QUB_TreeMonoIter segi = dataSet.find("Segment"); ! segi->isNull(); segi.nextSameName())
		++nseg;

	int iseg = 0;
	for (QUB_TreeMonoIter segi = dataSet.find("Segment"); ! segi->isNull(); segi.nextSameName(), ++iseg) {
		if ( StopFlag )
			break;
		if ( config["Tension"].dataAsInt() )
			pager = new TensionPulses_Pager(*segi, config, fillDataCB, report, pctCB);
		else if ( config["SpringDashPot"].dataAsInt() )
			pager = new SpringDashPot_Pager(*segi, fillDataCB, IDLCOND_PAGESIZE, report, -latency,
										    sampling*1e-3, config["SDP Damping"].dataAsDouble(0.6),
											config["SDP Freq"].dataAsDouble(1100.0));
		else
			pager = new LoadDataCB_Pager(*segi, fillDataCB, IDLCOND_PAGESIZE, report, - latency);

		vector<double> pageAmps(pager->pageSize);
		vector<int> pageAmpAssigned(pager->pageSize);
		for ( int pagestart=0; pagestart < pager->dataCount; pagestart += pager->pageSize ) {
			if ( StopFlag )
				break;
			pct = (int) floor(iseg*80.0/nseg + pagestart*80.0/pager->dataCount/nseg);
			pctCB->NodeCall( pctNode );

			pager->nextPage();
			double *data = pager->data;
			int pageBound = IMIN(pager->dataCount, pagestart + pager->pageSize) - pagestart;

			// set pageAmps to nearest known, with pageAmpAssigned telling if it's within delta
			if ( known.size() ) {
				for ( int i=0; i<pageBound; ++i ) {
					pageAmpAssigned[i] = FindKnownAmp(known, data[i], delta, pageAmps[i]);
				}
			}

			// set the rest of the pageAmps to delta-separated levels if requested
			if ( addDeltas ) {
				count = 0;
				for ( int i=0; i < pageBound; ++i ) {

					// end-of-event ?
					if ( (count==0) || pageAmpAssigned[i] || (abs(amp - data[i]) >= delta) ) {
						// assign pageAmps for last event
						if ( count ) {
							double avg = 0.0;
							for ( int j=count; j>0; --j )
								avg += data[i-j];
							avg /= count;
							for ( int j=count; j>0; --j )
								pageAmps[i-j] = avg;
						}

						// setup next event, ignoring assigned (known)
						if ( pageAmpAssigned[i] )
							count = 0;
						else {
							count = 1;
							amp = data[i];
						}
					}
					else {
						++count;
					}
				}
				// assign pageAmps for last event (last count points of page)
				if ( count ) {
					double avg = 0.0;
					for ( int j=count; j>0; --j )
						avg += data[pageBound-j];
					avg /= count;
					for ( int j=count; j>0; --j )
						pageAmps[pageBound-j] = avg;
				}
			} // addDeltas

			// generate events
			amp = pageAmps[0];
			count = 1;
			for ( int i=1; i<pageBound; ++i ) {
				if ( amp != pageAmps[i] ) {
					eventMaker.add(amp, count);
					amp = pageAmps[i];
					count = 1;
				}
				else {
					++count;
				}
			}
			if ( count ) {
				eventMaker.add(amp, count);
			}
		} // page

		if ( pager ) {
			delete pager;
			pager = NULL;
		}
	} // segment
	
	if ( pager )
		delete pager;

	pct = 80;
	pctCB->NodeCall(pctNode);

	eventMaker.consolidate(delta, known, StopFlag);

	pct = 90;
	pctCB->NodeCall(pctNode);

	// build output segments
	QUB_Tree result = QUB_Tree::Create("DataSet");
	QUB_TreeIter ins_result = result.end();
	for (QUB_TreeMonoIter tri = dataSet.children(); !tri->isNull(); ++tri) {
		if ( StopFlag )
			break;
		if ( tri->name() == "Segment" ) {
			int rem = tri->dataAs(1, 0, 0) - tri->dataAs(0, 0, 0) + 1;
			vector<int> segCounts, segClasses;
			int cls;
			while ( rem ) {
				eventMaker.nextEvent(cls, count, rem);
				if ( ! count )
					break; // should never happen because there should be exactly enough sum count for all segments
				rem -= count;
				segClasses.push_back( cls );
				segCounts.push_back( count );
			}
			ins_result.insert( CreateIdealSegment(*tri, segClasses, segCounts, eventMaker.ampOfCls, sampling) );
			++ins_result;
		}
		else {
			// blind-copy non-Segment subtrees of DataSet
			ins_result.insert(tri->clone());
			++ins_result;
		}
	}

	result = ApplyDeadTime(result, config["MinDur_MS"].dataAsDouble() * .9999999); // slightly smaller so as not to catch events of exactly MinDurMS

	pct = 100;
	pctCB->NodeCall(pctNode);

	return result;
}

QUB_Tree IdlCond( vector<QUB_Tree>& dataSets, QUB_Tree config, QTR_NodeFptr fillDataCB, 
				  QTR_NodeFptr reportCB, QTR_NodeFptr pctCB )
{
	mil_reportstream report(reportCB);
	QUB_Tree result = QUB_Tree::Create("IdlCond Result");
	result.appendClone(config);
	QUB_Tree resultDataSets = result.appendChild("DataSets");
	for ( vector<QUB_Tree>::iterator dsi = dataSets.begin(); dsi != dataSets.end(); ++dsi )
		resultDataSets.appendChild( IdlCond1DataSet(*dsi, config, fillDataCB, report, pctCB ) );
	return result;
}

//----------------------------------------------------------------------------
extern "C" MIL_API QTR_Impl * __stdcall
IdlCond_iface( QTR_Impl *cfg, QTR_Impl **data, QTR_INodeCFunc *fillDataCB, 
			  QTR_INodeCFunc *rptCB, QTR_INodeCFunc *pctCB ) {
	QUB_Tree cfgNode( cfg );

	vector<QUB_Tree> dataNodes;
	QTR_Impl **di = data;
	while ( *di )
		dataNodes.push_back( QUB_Tree( *(di++) ) );

	QTR_NodeFptr fillDat, rpt, pct;
	if ( fillDataCB )
		fillDat = QTR_NodeFptr( new QTR_NodePasFunc( fillDataCB ) );
	else
		fillDat = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( rptCB )
		rpt = QTR_NodeFptr( new QTR_NodePasFunc( rptCB ) );
	else
		rpt = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( pctCB )
		pct = QTR_NodeFptr( new QTR_NodePasFunc( pctCB ) );
	else
		pct = QTR_NodeFptr( new QTR_NoNodeFunc );

	QUB_Tree resultNode = IdlCond( dataNodes, cfgNode, fillDat, rpt, pct );
	
	QTR_Impl *result = resultNode.getImpl();
	QTR_INCREF( result );
	return result;
	}



//------------------------------------------


class EvtGen
{
public:
	virtual ~EvtGen() {}
	virtual bool hasNext() = 0;
	virtual void next(int& cls, int& count, double& amp) = 0;
};

class RangeOfIdlEvtGen : public EvtGen
{
	QUB_IdlSegments::iterator iter;
	int first, last;
public:
	RangeOfIdlEvtGen(QUB_IdlSegments::iterator& i_iter, int i_first, int i_last)
		: iter(i_iter), first(i_first), last(i_last)
	{}
	virtual ~RangeOfIdlEvtGen()
	{
		first = last = 0; // compiler please fully destroy this class
		// the preceding line is apparently crucial -- otherwise iter's destructor is never called,
		// so idlsegments is left locked (by the task thread) and the gui thread hangs when it tries to draw idl
		// empty destructor didn't work.
	}
	virtual bool hasNext()
	{
		return ( iter && (iter.segfirst() + iter->first) <= last );
	}
	virtual void next(int& cls, int& count, double& amp)
	{
		int f, l;
		iter.read(&f, &l, &cls, &amp);
		count = IMIN(l, last) - IMAX(f, first) + 1;
		++iter;
	}
};

class SegNodeIdlGen : public EvtGen
{
	QUB_Tree segment;
	int ndwt, i;
	int *idwt, *cdwt;
	double *amps;
public:
	SegNodeIdlGen(QUB_Tree i_segment)
		: segment(i_segment)
	{
		ndwt = segment["DwellCount"].dataAsInt();
		idwt = (int*) segment["Classes"].data();
		cdwt = (int*) segment["Counts"].data();
		amps = (double*) segment["amp"].data();
		i = 0;
	}
	virtual ~SegNodeIdlGen()
	{
		i = 0; // compiler please destroy properly
	}
	virtual bool hasNext()
	{
		return (i < ndwt);
	}
	virtual void next(int& cls, int& count, double& amp)
	{
		cls = idwt[i];
		count = cdwt[i];
		amp = amps[cls];
		++i;
	}
};

class JoinEvtGen : public EvtGen
{ // joins adjacent events of same class
	EvtGen *gen;
	bool haveNext;
	int nextCls, nextCount;
	double nextAmp;
public:
	JoinEvtGen(EvtGen *evtGen)
		: gen(evtGen)
	{
		if ( haveNext = gen->hasNext() )
			gen->next(nextCls, nextCount, nextAmp);
	}
	virtual ~JoinEvtGen()
	{
		nextCls = 0; // compiler please destroy this fully
	}
	virtual bool hasNext()
	{
		return ( haveNext );
	}
	virtual void next(int& cls, int& count, double& amp)
	{
		cls = nextCls;
		amp = nextAmp;
		count = 0;
		do {
			count += nextCount;
			if ( haveNext = gen->hasNext() )
				gen->next(nextCls, nextCount, nextAmp);
		} while (haveNext && (nextCls == cls));
	}
};

class MplexEvtGen : public EvtGen
{
	vector<EvtGen*> gens;
	vector< pair<int, int> > nexts; // (cls, count) // indicate !havenext by count=0?
	vector< double > nextAmps;
	bool haveNext;
	map< vector<int>, int > clsOfGroup; // this depends on vectors having operator< defined

public:
	vector< vector<int> > classes;
	vector< vector<double> > amps;

	MplexEvtGen()
		: haveNext(true)
	{}
	MplexEvtGen(QUB_Tree mclasses, QUB_Tree mamps)
		: haveNext(true)
	{
		if ( ! mclasses.isNull() ) {
			for (int i=0; i<(int)mclasses.dataRows(); ++i) {
				vector<int> mclass(mclasses.dataCols());
				vector<double> mamp(mclasses.dataCols());
				for (int j=0; j<(int)mclass.size(); ++j) {
					mclass[j] = mclasses.dataAs(i, j, 0);
					mamp[j]   = mamps.dataAs(i, j, 0.0);
				}
				classes.push_back(mclass);
				amps.push_back(mamp);
				clsOfGroup[mclass] = i;
			}
		}
	}
	virtual ~MplexEvtGen()
	{
		haveNext = false; // compiler please destroy this? (it's being used on the stack so maybe not an issue here?)
	}
	void add(EvtGen *gen)
	{
		gens.push_back(gen);
		int ix = nexts.size();
		nexts.resize(ix+1);
		nextAmps.resize(ix+1);
		getNext(ix);
	}
	virtual bool hasNext()
	{
		return haveNext;
	}
	virtual void next(int& cls, int& count, double& amp)
	{
		cls = findOrAddCls();
		int minCount = nexts[0].second;
		for ( int i=nexts.size()-1; i>0; --i )
			if ( nexts[i].second < minCount )
				minCount = nexts[i].second;
		count = minCount;

		for ( int i=nexts.size()-1; i>=0; --i ) {
			nexts[i].second -= minCount;
			if ( nexts[i].second == 0 )
				getNext(i);
		}
	}

	QUB_Tree getMplexClasses()
	{
		QUB_Tree mplexClasses = QUB_Tree::Create("MplexClasses");
		mplexClasses.setNumData(QTR_TYPE_INT, classes.size(), gens.size(), (void**)0);
		for ( int c=0; c<(int)classes.size(); ++c ) {
			vector<int>& mcls = classes[c];
			for ( int i=0; i<(int)mcls.size(); ++i )
				mplexClasses.dataAs(c, i, 0) = mcls[i];
		}
		return mplexClasses;
	}

	QUB_Tree getMplexAmps()
	{
		QUB_Tree mplexAmps = QUB_Tree::Create("MplexAmps");
		mplexAmps.setNumData(QTR_TYPE_DOUBLE, amps.size(), gens.size(), (void**)0);
		for ( int c=0; c<(int)amps.size(); ++c ) {
			vector<double>& mamp = amps[c];
			for ( int i=0; i<(int)mamp.size(); ++i )
				mplexAmps.dataAs(c, i, 0.0) = mamp[i];
		}
		return mplexAmps;
	}

private:
	void getNext(int ix)
	{
		if ( gens[ix]->hasNext() )
			gens[ix]->next(nexts[ix].first, nexts[ix].second, nextAmps[ix]);
		else {
			nexts[ix].second = 0;
			haveNext = false;
		}
	}
	int findOrAddCls() // of current crop of nexts
	{
		vector<int> mclass(nexts.size());
		vector<double> mamp(nexts.size());
		for ( int i=nexts.size()-1; i>=0; --i ) {
			mclass[i] = nexts[i].first;
			mamp[i]   = nextAmps[i];
		}
		map< vector<int>, int >::iterator mci = clsOfGroup.find(mclass);
		if ( mci != clsOfGroup.end() ) {
			return mci->second;
		}
		else {
			int ix = classes.size();
			clsOfGroup[mclass] = ix;
			classes.push_back(mclass);
			amps.push_back(mamp);
			return ix;
		}
	}

};

// breaks each event up into stretches that are a power of 2 points long
// the length of the output event in points is pow(2, event.count).
// also, as per MIP, where the output contains a stretch of events of the same class,
// the first and last events will always have a count of 1
class Pow2EvtGen : public EvtGen
{
	EvtGen *gen;
	int maxCount;
	int xtraCls, xtraCount;
	double xtraAmp;
public:
	Pow2EvtGen(EvtGen *evtGen, int maxPower)
		: gen(evtGen), maxCount(1<<maxPower), xtraCount(0)
	{}
	virtual bool hasNext()
	{
		return ( xtraCount || gen->hasNext() );
	}
	virtual void next(int& cls, int& count, double& amp)
	{
		if ( xtraCount == 0 ) { // get next event, return the first point
			gen->next(xtraCls, xtraCount, xtraAmp);
			cls = xtraCls;
			count = 0;
			amp = xtraAmp; //?
			xtraCount -= 1;
		}
		else if ( xtraCount == 1 ) { // the algorithm below fails for count<2 [this is the last point that that algorithm is trying to avoid]
			cls = xtraCls;
			count = 0;
			amp = xtraAmp; //?
			xtraCount = 0;
		}
		else {
			cls = xtraCls;
			int p2, p2count;
			for ( p2=0, p2count=1; p2count<xtraCount; ++p2, p2count<<=1 )
				; // exits with p2count at the first power of 2 >= xtraCount
			do {  // we want to split off the last point, so we go one power lower,
				--p2;
				p2count >>= 1;
			} while (p2count > maxCount); // and continue dropping until we're at or below max power
			count = p2;
			amp = xtraAmp; //?
			xtraCount -= p2count;
		}
	}
};

QUB_Tree MplexIdl_none(vector<int>& firsts, vector<int>& lasts)
{
	QUB_Tree dataSet = QUB_Tree::Create("DataSet");
	QUB_TreeIter dataSet_ins = dataSet.end();

	for (int segi = 0; segi < (int)firsts.size(); ++segi) {
		dataSet_ins.insert("Segment");
		QUB_Tree segment = *dataSet_ins;
		++dataSet_ins;

		segment.setNumData(QTR_TYPE_INT, 2, 1, (void*)0);
		segment.dataAs(0, 0, 0) = firsts[segi];
		segment.dataAs(1, 0, 0) = lasts[segi];

		// one dwell whole segment class 0 multiclass:[]
		segment["DwellCount"].setData(QTR_TYPE_INT, 1);
		segment["Classes"].setData(QTR_TYPE_INT, 0);
		segment["Counts"].setData(QTR_TYPE_INT, lasts[segi] - firsts[segi] + 1);
		segment["amp"].setData(QTR_TYPE_DOUBLE, 0.0);
	}

	dataSet["MplexClasses"]; // 1 row of 0 columns // == no data
	dataSet["MplexAmps"];

	return dataSet;
}

QUB_Tree MplexIdl(vector<QUB_IdlSegments*>& idls, vector<int>& firsts, vector<int>& lasts)
{
	if ( idls.size() == 0 )
		return MplexIdl_none(firsts, lasts);

	QUB_Tree dataSet = QUB_Tree::Create("DataSet");
	QUB_TreeIter dataSet_ins = dataSet.end();

	QUB_Tree mplexClasses; // a segment's classes are used to initialize the next segment, so all share the same classes
	QUB_Tree mplexAmps;

	for (int segi = 0; segi < (int)firsts.size(); ++segi) {
		vector<RangeOfIdlEvtGen*> ranged;
		vector<JoinEvtGen*> joined;

		dataSet_ins.insert("Segment");
		QUB_Tree segment = *dataSet_ins;
		++dataSet_ins;

		segment.setNumData(QTR_TYPE_INT, 2, 1, (void*)0);
		segment.dataAs(0, 0, 0) = firsts[segi];
		segment.dataAs(1, 0, 0) = lasts[segi];

		try {
			for ( int i=0; i<(int)idls.size(); ++i )
				ranged.push_back( new RangeOfIdlEvtGen(idls[i]->find(firsts[segi]), firsts[segi], lasts[segi]) );
			for ( int i=0; i<(int)idls.size(); ++i )
				joined.push_back( new JoinEvtGen(ranged[i]) );
			MplexEvtGen mplex(mplexClasses, mplexAmps);
			for ( int i=0; i<(int)idls.size(); ++i )
				mplex.add(joined[i]);

			vector<int> clss, counts;
			int cls, count;
			int maxcls = -1;
			double amp; // ignored -- amp of mplexed class is undefined
			while ( mplex.hasNext() ) {
				mplex.next(cls, count, amp);
				clss.push_back(cls);
				counts.push_back(count);
				maxcls = max(maxcls, cls);
			}

			mplexClasses = mplex.getMplexClasses();
			mplexAmps    = mplex.getMplexAmps();
			
			segment["DwellCount"].setData(QTR_TYPE_INT, clss.size());
			segment["Classes"].setNumData(QTR_TYPE_INT, clss.size(), 1, (int*)(&clss[0]));
			segment["Counts"].setNumData(QTR_TYPE_INT, counts.size(), 1, (int*)(&counts[0]));
			segment["amp"].setNumData(QTR_TYPE_DOUBLE, maxcls+1, 1, (void*)0);
		}
		catch (...) {
		}
		for ( int i=0; i<(int)joined.size(); ++i )
			delete joined[i];
		for ( int i=0; i<(int)ranged.size(); ++i )
			delete ranged[i];
	}

	dataSet.insertChild(mplexClasses);
	dataSet.insertChild(mplexAmps);

	return dataSet;
}

extern "C" MIL_API QTR_Impl* __stdcall
MplexIdl_iface(QUB_IdlSegments** idlarr, int nseg, int* firsts, int* lasts)
{
	vector<QUB_IdlSegments*> idls;
	for ( ; *idlarr; ++idlarr )
		idls.push_back( *idlarr );
	vector<int> vfirsts, vlasts;
	for (int i=0; i<nseg; ++i) {
		vfirsts.push_back(firsts[i]);
		vlasts.push_back(lasts[i]);
	}
	QUB_Tree result = MplexIdl(idls, vfirsts, vlasts);
	QTR_Impl* resultPtr = result.getImpl();
	QTR_INCREF( resultPtr );
	return resultPtr;
}

QUB_Tree MplexIdlWithCond(QUB_IdlSegments *idl, QUB_Tree condIdl, int maxPower)
{
	QUB_Tree dataSet = QUB_Tree::Create("DataSet");
	QUB_TreeIter dataSet_ins = dataSet.end();

	QUB_Tree mplexClasses = condIdl["MplexClasses"];
	dataSet_ins.insert(mplexClasses.clone());
	++dataSet_ins;

	dataSet_ins.insert(condIdl["MplexAmps"].clone());
	++dataSet_ins;

	for (QUB_TreeMonoIter segi = condIdl.find("Segment"); ! segi->isNull(); segi.nextSameName()) {
		QUB_Tree segment = segi->clone(false);
		dataSet_ins.insert(segment);
		++dataSet_ins;

		RangeOfIdlEvtGen events(idl->find(segment.dataAs(0,0,0)), segment.dataAs(0,0,0), segment.dataAs(1,0,0));
		JoinEvtGen joinedEvents(&events);
		SegNodeIdlGen condEvents(*segi);
		MplexEvtGen mplex;
		mplex.add(&joinedEvents);
		mplex.add(&condEvents);
		Pow2EvtGen pow2gen(&mplex, maxPower);

		vector<int> clss, counts;
		int cls, count;
		double amp;
		while ( pow2gen.hasNext() ) {
			pow2gen.next(cls, count, amp);
			clss.push_back(cls);
			counts.push_back(count);
		}

		int ndwt = counts.size();
		QUB_Tree biplex = mplex.getMplexClasses();
		int *biplexClasses = (int*) biplex.data();
		// cls c : event cls biplex[2*c+0], cond cls biplex[2*c+1]
		
		segment["DwellCount"].setData(QTR_TYPE_INT, ndwt);
		segment["Powers"].setNumData(QTR_TYPE_INT, ndwt, 1, (int*)(&counts[0]));
		QUB_Tree classesNode = segment["Classes"];
		classesNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
		int *classes = (int*) classesNode.data();
		QUB_Tree condClassesNode = segment["CondClasses"];
		condClassesNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
		int *condClasses = (int*) condClassesNode.data();

		for ( int i=0; i<ndwt; ++i ) {
			classes[i] = biplexClasses[2*clss[i]+0];
			condClasses[i] = biplexClasses[2*clss[i]+1];
		}
	}

	return dataSet;
}

extern "C" MIL_API QTR_Impl* __stdcall
MplexIdlWithCond_iface(QUB_IdlSegments* idl, QTR_Impl *condIdl, int maxPower)
{
	QUB_Tree result = MplexIdlWithCond(idl, QUB_Tree(condIdl), maxPower);
	QTR_Impl* resultPtr = result.getImpl();
	QTR_INCREF( resultPtr );
	return resultPtr;
}


